"""
Module for Roles
https://docs.cloud.f5.com/docs/api/role
"""
from uplink import Consumer, Path, Body, json, get, post, put, delete # pylint: disable=unused-import
from . import helper


@helper.common_decorators
class Role(Consumer):
    """
    Class for Roles
    "custom" endpoints in use for list(), get(), create()
    """
    def __init__(self, session):
        super().__init__(base_url=session._tenant_url, client=session._session)

    @get('/api/web/custom/namespaces/{namespace}/roles')
    def list(self, namespace: Path ='system'):
        """List all Roles"""

    @get('/api/web/namespaces/{namespace}/roles')
    def listv1(self, namespace: Path = 'system'):
        """List all Roles"""

    @json
    @post('/api/web/custom/namespaces/{namespace}/roles')
    def create(self, payload: Body, namespace: Path = 'system'):
        """
        Create a Role
        Use create_payload() to build Body
        """

    @json
    @put('/api/web/custom/namespaces/{namespace}/roles/{name}')
    def replace(self, payload: Body, name: Path, namespace: Path = 'system'):
        """
        Replace Role
        Use replace_payload to build Body
        """

    @json
    @delete('/api/web/namespaces/{namespace}/roles/{name}')
    def delete(self, payload: Body, name: Path, namespace: Path = 'system'):
        """
        Delete a Role
        use delete_payload() to build Body
        ISSUE: Console shows no payload here
        """

    @staticmethod
    def create_payload(
            name: str,
            description: str,
            api_groups: list,
            labels = None
        ):
        """Payload for Create"""
        if labels is None:
            labels = {}
        return {
            "api_groups": api_groups,
            "metadata": {
                "annotations": {},
                "description": description,
                "disable": False,
                "labels": labels,
                "name": name,
                "namespace": "system"
            },
            "namespace": "system",
            "spec": {}
        }

    @staticmethod
    def replace_payload(name: str, api_groups: list):
        """
        Payload for Role modifications
        *api_groups is being replaced
        """
        return {
            "api_groups": api_groups,
            "name": name,
            "namespace": "system",
            "spec": {}
        }

    @staticmethod
    def delete_payload(name: str):
        """Payload for delete"""
        return {
            "fail_if_referred": True,
            "name": name,
            "namespace": "system"
        }



"""
Module for Load Balancers
https://docs.cloud.f5.com/docs-v2/api/views-http-loadbalancer
https://docs.cloud.f5.com/docs-v2/api/views-tcp-loadbalancer
"""
from uplink import Consumer, Path, Body, get, post, put, delete, json
from . import helper


@helper.common_decorators
class HTTPLoadBalancer(Consumer):
    """
    Class for HTTP Load Balancer API.
    """

    def __init__(self, session):
        """
        Initialize the HTTPLoadBalancer Consumer.
        :param session: Session object with tenant URL and auth.
        """
        super().__init__(base_url=session._tenant_url, client=session._session)

    @get('/api/config/namespaces/{namespace}/http_loadbalancers')
    def list(self, namespace: Path):
        """
        List all HTTP Load Balancers in a namespace.
        """

    @get('/api/config/namespaces/{namespace}/http_loadbalancers/{name}')
    def get(self, namespace: Path, name: Path):
        """
        Get details of an HTTP Load Balancer.
        """

    @json
    @post('/api/config/namespaces/{namespace}/http_loadbalancers')
    def create(self, payload: Body, namespace: Path):
        """
        Create an HTTP Load Balancer.
        """

    @json
    @put('/api/config/namespaces/{namespace}/http_loadbalancers/{name}')
    def replace(self, payload: Body, namespace: Path, name: Path):
        """
        Replace an HTTP Load Balancer.
        """

    @delete('/api/config/namespaces/{namespace}/http_loadbalancers/{name}')
    def delete(self, namespace: Path, name: Path):
        """
        Delete an HTTP Load Balancer.
        """

@helper.common_decorators
class TCPLoadBalancer(Consumer):
    """
    Class for TCP Load Balancer API.
    """

    def __init__(self, session):
        """
        Initialize the TCPLoadBalancer Consumer.
        :param session: Session object with tenant URL and auth.
        """
        super().__init__(base_url=session._tenant_url, client=session._session)

    @get('/api/config/namespaces/{namespace}/tcp_loadbalancers')
    def list(self, namespace: Path):
        """
        List all TCP Load Balancers in a namespace.
        """

    @get('/api/config/namespaces/{namespace}/tcp_loadbalancers/{name}')
    def get(self, namespace: Path, name: Path):
        """
        Get details of a TCP Load Balancer.
        """

    @json
    @post('/api/config/namespaces/{namespace}/tcp_loadbalancers')
    def create(self, payload: Body, namespace: Path):
        """
        Create a TCP Load Balancer.
        """

    @json
    @put('/api/config/namespaces/{namespace}/tcp_loadbalancers/{name}')
    def replace(self, payload: Body, namespace: Path, name: Path):
        """
        Replace a TCP Load Balancer.
        """

    @delete('/api/config/namespaces/{namespace}/tcp_loadbalancers/{name}')
    def delete(self, namespace: Path, name: Path):
        """
        Delete a TCP Load Balancer.
        """


"""Module for Namespaces"""
from uplink import Consumer, Path, Body, json, get, post # pylint: disable=unused-import
from . import helper


@helper.common_decorators
class NS(Consumer):
    """Class for Namespaces"""
    def __init__(self, session):
        super().__init__(base_url=session._tenant_url, client=session._session)

    @get('/api/web/namespaces')
    def list(self):
        """List all Namespaces"""

    @get('/api/web/namespaces/{name}')
    def get(self, name: Path):
        """Get a single Namespace"""

    @json
    @post('/api/web/namespaces')
    def create(self, payload: Body):
        """
        Create a Namespace
        Use create_payload() to build Body
        """

    @json
    @post('/api/web/namespaces/{name}/cascade_delete')
    def delete(self, payload: Body, name: Path):
        """
        Cascade delete a Namespace
        use delete_payload() to build Body
        """

    @staticmethod
    def create_payload(name: str, description: str):
        """Payload for create"""
        return {
            'metadata': {
                'annotations': {},
                'description': description,
                'disable': False,
                'labels': {},
                'name': name,
                'namespace': ''
            },
            'spec': {}
        }

    @staticmethod
    def delete_payload(name: str):
        """Payload for delete"""
        return {
            'name:': name
        }
    


"""
Module for Tenant Methods
https://docs.cloud.f5.com/docs/api/tenant
"""
from datetime import datetime
from uplink import Consumer, QueryMap, Path, Body, json, get, post, put, delete # pylint: disable=unused-import
from . import helper

@helper.common_decorators
class Tenant(Consumer):
    """
    Class for Tenant Methods
    """
    def __init__(self, session):
        super().__init__(base_url=session._tenant_url, client=session._session)

    @get('/api/web/namespaces/system/tenant/idm/events/last_login')
    def last_login(self):
        """Lists last login for each user"""

    @get('/api/web/namespaces/system/tenant/idm/events/login')
    def login_events(self, **params: QueryMap):
        """
        Lists login events
        params:
          first: (int) offset
          max: (int) records in response
        """

    @get('api/web/namespaces/system/tenant/idm/users/inactive')
    def list_inactive_users(self):
        """Lists inactive users (90 days without login)"""

    @get('/api/web/namespaces/system/tenant/settings')
    def get_settings(self):
        """Gets tenant settings"""

    @get('/api/web/namespaces/system/tenant/idm/settings')
    def get_idm_settings(self):
        """Gets tenant IDM settings"""

    @json
    @post('/api/web/namespaces/system/tenant/idm/events/login_in_time')
    def login_events_in_tf(self, payload: Body):
        """
        Login events for specified time frame
        Use login_events_in_tf_payload() to build Body
        """

    @staticmethod
    def login_events_in_tf_payload(start: datetime, end: datetime, first: int = 0, maximum: int = 0): # pylint: disable=line-too-long
        """Payload for login_events_in_tf"""
        return {
            "end": helper.xc_format_date(end),
            "first": first,
            "max": maximum,
            "start": helper.xc_format_date(start)
        }



"""
Module for Origin Pool
https://docs.cloud.f5.com/docs-v2/api/views-origin-pool
"""
from uplink import Consumer, Path, Body, get, post, put, delete, json  # pylint: disable=unused-import
from . import helper


@helper.common_decorators
class OriginPool(Consumer):
    """
    Class for interacting with the Origin Pool API.
    """

    def __init__(self, session):
        """
        Initialize the OriginPool Consumer.
        :param session: Session object with tenant URL and authentication.
        """
        super().__init__(base_url=session._tenant_url, client=session._session)

    @get('/api/config/namespaces/{namespace}/origin_pools')
    def list(self, namespace: Path):
        """
        List all Origin Pools in a namespace.
        """

    @get('/api/config/namespaces/{namespace}/origin_pools/{name}')
    def get(self, namespace: Path, name: Path):
        """
        Get details of a specific Origin Pool.
        """

    @json
    @post('/api/config/namespaces/{namespace}/origin_pools')
    def create(self, payload: Body, namespace: Path):
        """
        Create an Origin Pool.
        The payload must be provided by the user.
        """

    @json
    @put('/api/config/namespaces/{namespace}/origin_pools/{name}')
    def replace(self, payload: Body, namespace: Path, name: Path):
        """
        Replace an Origin Pool.
        The payload must be provided by the user.
        """

    @delete('/api/config/namespaces/{namespace}/origin_pools/{name}')
    def delete(self, namespace: Path, name: Path):
        """
        Delete an Origin Pool.
        """


"""Package level imports""" # pylint: disable=invalid-name
from .session import Session as session
from .ns import NS as ns
from .cred import APIcred as apicred
from .cred import SVCcred as svccred
from .user import User as user
from .namespace_role import NSrole as nsrole
from .group import Group as group
from .role import Role as role
from .tenant import Tenant as tenant
from .registration import Registration as registration
from .xcsite import Site as site
from .cert import Cert as cert
from .origin_pool import OriginPool as origin_pool
from .load_balancer import HTTPLoadBalancer as http_loadbalancer
from .load_balancer import TCPLoadBalancer as tcp_loadbalancer



"""
Module for XC Site Methods
https://docs.cloud.f5.com/docs/api/site
"""
from uplink import Consumer, Path, Body, Query, json, get, post #pylint: disable=unused-import
from . import helper

@helper.common_decorators
class Site(Consumer):
    """
    Class for Site Methods
    """
    def __init__(self, session):
        super().__init__(base_url=session._tenant_url, client=session._session)

    @get('/api/config/namespaces/{namespace}/sites')
    def list(self, namespace: Path ='system', report_fields: Query = None, report_status_fields: Query = None ): #pylint: disable=line-too-long
        """List Sites"""

    @get('/api/config/namespaces/{namespace}/sites/{name}')
    def get(self, name: Path, namespace: Path ='system'):
        """Get a single Site"""

    @json
    @post('/api/config/namespaces/{namespace}/sites/{name}/upgrade_os')
    def upgrade_os(self, payload: Body, name: Path, namespace: Path ='system'):
        """
        Upgrade Site OS
        Use upgrade_payload() to build Body
        """

    @json
    @post('/api/config/namespaces/{namespace}/sites/{name}/upgrade_sw')
    def upgrade_sw(self, payload: Body, name: Path, namespace: Path ='system'):
        """
        Upgrade Site Software
        Use upgrade_payload() to build Body
        """

    @json
    @post('/api/register/namespaces/{namespace}/site/{name}/state')
    def state(self, payload: Body, name: Path, namespace: Path ='system'):
        """
        Update site state
        Use state_payload() to build Body
        """

    @staticmethod
    def upgrade_payload(name: str, version: str, namespace: str = 'system') -> dict:
        """
        Payload for upgrade_*
        current_dw = "crt-20240329-2728"
        current_os = "9.2024.6" or "7.2009.45" (depending on RHEL vs. CentOS)
        """
        return {
            "name": name,
            "namespace": namespace,
            "version": version
        }

    @staticmethod
    def state_payload(name: str, namespace: str = 'system', state = 'DECOMMISSIONING') -> dict:
        """
        Payload for decom_site
        """
        return {
            "name": name,
            "namespace": namespace,
            "state": state
        }



"""
Module for Namespace Roles
https://docs.cloud.f5.com/docs/api/namespace-role
Need to unwind this class -- not useful
"""
from uplink import Consumer, Path, Body, json, get, post, put # pylint: disable=unused-import
from . import helper


@helper.common_decorators
class NSrole(Consumer):
    """Class for Namespace Roles"""
    def __init__(self, session):
        super().__init__(base_url=session._tenant_url, client=session._session)

    @get('/api/web/namespaces/{namespace}/namespace_roles')
    def list(self, namespace: Path = 'system'):
        """List all Namespace Roles"""

    @get('/api/web/namespaces/{namespace}/namespace_roles/{name}')
    def get(self, namespace: Path, name: Path,):
        """Get a single Namespace Role"""

    @json
    @post('/api/web/namespaces/{namespace}/namespace_roles')
    def create(self, payload: Body, namespace: Path):
        """
        Create a Namespace Role
        Use create_payload() to build Body
        """

    @json
    @post('/api/web/custom/namespaces/{namespace}/user_groups/{name}')
    def delete(self, name: Path, namespace: Path = 'system'):
        """Delete a group"""

    @json
    @put('/api/web/custom/namespaces/{namespace}/user_groups/{name}/assign_namespace_roles')
    def role_assign(self, payload: Body, name: Path, namespace: Path = 'system'):
        """
        Assign roles for a group
        Use role_payload() to build Body
        """

    @json
    @put('/api/web/custom/namespaces/{namespace}/user_groups/{name}/remove_namespace_roles')
    def role_remove(self, payload: Body, name: Path, namespace: Path = 'system'):
        """
        Remove roles for a group
        Use role_payload() to build Body
        """

    @staticmethod
    def create_payload(
            name: str,
            namespace: str,
            user: list,
            role: list
        ):
        """Payload for Create"""
        return {
            "metadata": {
                "name": name,
                "namespace": namespace
            },
            "spec": {
                "user": user,
                "role": role
            }
        }

    @staticmethod
    def role_payload(
            name: str,
            namespace_roles: list
        ):
        """
        Payload for role modifications
        *namespace_roles are added/deleted based on method
        """
        return {
            "name": name,
            "namespace_roles": namespace_roles
        }



"""
Module for Registration Methods
https://docs.cloud.f5.com/docs/api/registration
"""
from uplink import Consumer, Path, Body, json, get, post #pylint: disable=unused-import
from . import helper

@helper.common_decorators
class Registration(Consumer):
    """
    Class for Tenant Methods
    """
    def __init__(self, session):
        super().__init__(base_url=session._tenant_url, client=session._session)

    @get('/api/register/namespaces/{namespace}/registrations')
    def list(self, namespace: Path ='system'):
        """List Registrations"""

    @json
    @post('/api/register/namespaces/{namespace}/listregistrationsbystate')
    def list_by_state(self, payload: Body, namespace: Path ='system'):
        """
        List Registrations by state
        Use list_by_state_payload() to build Body
        """

    @get('/api/register/namespaces/{namespace}/registrations/{name}')
    def get(self, name: Path, namespace: Path ='system'):
        """Get a single Registration"""

    @json
    @post('/api/register/namespaces/{namespace}/registration/{name}/approve')
    def approve(self, payload: Body, name: Path, namespace: Path ='system'):
        """
        Approve a pending registration
        Use approve_payload() to build Body
        """

    @staticmethod
    def list_by_state_payload(state: str = 'NEW', namespace: str = 'system') -> dict:
        """
        Payload for list_by_state
        """
        return {
            "namespace": namespace,
            "state": state
        }

    @staticmethod
    def approve_payload(name: str, passport: dict, state: str = 'APPROVED', namespace: str = 'system') -> dict: # pylint: disable=line-too-long
        """
        Payload for approve
        pass 'passport' dict from get()
        """
        return {
            "namespace": namespace,
            "name": name,
            "state": state,
            "passport" : passport
    }

    @staticmethod
    def _get_reg_name(obj: dict) -> str:
        """
        Takes a registration object, returns the name
        """
        return obj['name']

    @staticmethod
    def _get_reg_token(obj: dict) -> str:
        """
        Takes a registration object, returns the name
        """
        return obj['get_spec']['token']

    @staticmethod
    def _get_passport(obj: dict) -> dict:
        """
        Takes a registration object, returns the passport
        """
        return obj['get_spec']['passport']

    @staticmethod
    def _get_cluster_name(obj: dict) -> str:
        """
        Takes a registration object, returns the passport
        """
        return obj['get_spec']['passport']['cluster_name']



"""Module for Certificate"""
from uplink import Consumer, Path, Body, json, get, post, put, delete #pylint: disable=unused-import
from . import helper


@helper.common_decorators
class Cert(Consumer):
    """Class for Certificates"""
    def __init__(self, session):
        super().__init__(base_url=session._tenant_url, client=session._session)

    @get('/api/config/namespaces/{namespace}/certificates')
    def list(self, namespace: Path):
        """List all Certificates in an NS"""

    @get('/api/config/namespaces/{namespace}/certificates/{name}')
    def get(self, name: Path, namespace: Path):
        """Get a Certificate"""

    @json
    @post('/api/config/namespaces/{namespace}/certificates')
    def create(self, payload: Body, namespace: Path):
        """Create a Certificate"""

    @json
    @put('/api/config/namespaces/{namespace}/certificates/{name}')
    def replace(self, payload: Body, name: Path, namespace: Path):
        """Replace a Certificate"""

    @json
    @delete('/api/config/namespaces/{namespace}/certificates/{name}')
    def delete(self, payload: Body, name: Path, namespace: Path):
        """Delete a Certificate"""

    @staticmethod
    def create_payload(name: str, namespace: str, cert: str, key: str):
        """
        Payload for create or replace cert
        {cert} and {key} must be b64 encoded
        """
        return {
            'metadata': {
                'name': name,
                'namespace': namespace,
                'disable': False
            },
            'spec': {
                'certificate_url': f'string:///{cert}',
                'private_key': {
                    'clear_secret_info': {
                        'url': f'string:///{key}'
                    }
                }
            }
        }

    @staticmethod
    def delete_payload(name: str, namespace: str, fail_if_referred: bool = True):
        """Payload for delete"""
        return {
            "fail_if_referred": fail_if_referred,
            "name": name,
            "namespace": namespace
        }



"""Module for API Credentials"""
from uplink import Consumer, Path, Body, json, get, post
from . import helper


@helper.common_decorators
class APIcred(Consumer):
    """Class for API Credentials"""
    def __init__(self, session):
        super().__init__(base_url=session._tenant_url, client=session._session)

    @get('/api/web/namespaces/{namespace}/api_credentials')
    def list(self, namespace: Path = 'system'):
        """List all API Credentials"""

    @get('/api/web/namespaces/{namespace}/api_credentials/{name}')
    def get(self, name: Path, namespace: Path = 'system'):
        """Get a single API Credential"""

    @json
    @post('/api/web/namespaces/{namespace}/api_credentials')
    def create(self, payload: Body, namespace: Path = 'system'):
        """
        Create an API Credential
        Use create_payload() to build Body
        """

    @json
    @post('api/web/namespaces/{namespace}/renew/api_credentials')
    def renew(self, payload: Body, namespace: Path = 'system'):
        """
        Renew an API Credential
        Use renew_payload() to build Body
        """

    @json
    @post('/api/web/namespaces/{namespace}/revoke/api_credentials')
    def revoke(self, payload: Body, namespace: Path = 'system'):
        """
        Revoke an API Credential
        Use revoke_payload() to build Body
        """

    @staticmethod
    def create_payload(name: str, expiration_days: int, namespace: str = 'system'):
        """Payload for create"""
        return {
            'spec': {
                'type': 'API_TOKEN'
            },
            'namespace': namespace,
            'name': name,
            'expiration_days': expiration_days
        }

    @staticmethod
    def renew_payload(name: str, expiration_days: int, namespace: str = 'system'):
        """Payload for renew"""
        return {
            'expiration_days': expiration_days,
            'name': name,
            'namespace': namespace
        }

    @staticmethod
    def revoke_payload(name: str, namespace: str = 'system'):
        """Payload for revoke """
        return {
            'name': name,
            'namespace': namespace
        }


@helper.common_decorators
class SVCcred(Consumer):
    """Class for Service Credentials"""
    def __init__(self, session):
        super().__init__(base_url=session._tenant_url, client=session._session)

    @get('/api/web/namespaces/{namespace}/service_credentials')
    def list(self, namespace: Path = 'system'):
        """List all Service Credentials"""

    @json
    @post('/api/web/namespaces/{namespace}/service_credentials')
    def create(self, payload: Body, namespace: Path = 'system'):
        """
        Create an Service Credential
        Use create_payload() to build Body
        """

    @json
    @post('api/web/namespaces/{namespace}/renew/service_credentials')
    def renew(self, payload: Body, namespace: Path = 'system'):
        """
        Renew a Service Credential
        Use renew_payload() to build Body
        """

    @json
    @post('/api/web/namespaces/{namespace}/revoke/service_credentials')
    def revoke(self, payload: Body, namespace: Path = 'system'):
        """
        Revoke a Service Credential
        Use revoke_payload() to build Body
        """

    @staticmethod
    def create_payload(
        name: str,
        namespace_roles: list,
        expiration_days: int,
        namespace: str = 'system'
        ):
        """Payload for create"""
        return {
            'type':'SERVICE_API_TOKEN',
            'namespace': namespace,
            'name': name,
            'namespace_roles': namespace_roles,
            'expiration_days': expiration_days
        }

    @staticmethod
    def renew_payload(name: str, expiration_days: int, namespace: str = 'system'):
        """payload for renew"""
        return {
            'expiration_days': expiration_days,
            'name': name,
            'namespace': namespace
        }

    @staticmethod
    def revoke_payload(name: str, namespace: str = 'system'):
        """Payload for revoke"""
        return {
            'name': name,
            'namespace': namespace
        }
    


"""Module providing XC session"""
from urllib.parse import urlparse
import requests
from . import helper

class Session:
    """Class providing request session with auth"""
    def __init__(self, tenant_url=None, api_token=None, validate=True):
        self._tenant_url = self.validate_url(tenant_url)
        self._api_token = api_token
        self._session = requests.Session()
        self._session.headers.update({'Authorization': f'APIToken {self._api_token}'})
        if validate:
            self.whoami()

    def whoami(self) -> None:
        """Method to check if we have a valid session"""
        try:
            r = self._session.get(self._tenant_url + '/api/web/custom/namespaces/system/whoami')
            r.raise_for_status()
        except Exception as e:
            raise helper.TopsXCException("Invalid Token") from e

    @staticmethod
    def validate_url(url) -> str:
        """Method to validate the tenant URL."""
        parsed_url = urlparse(url)
        if parsed_url.scheme and parsed_url.netloc:
            stripped_url = f"{parsed_url.scheme}://{parsed_url.netloc}{parsed_url.path.rstrip('/')}"
        else:
            raise helper.TopsXCException("Invalid Tenant URL")
        return stripped_url
        


"""
Module for Users
https://docs.cloud.f5.com/docs/api/user
"""
from uplink import Consumer, Path, Body, json, get, post, put
from . import helper


@helper.common_decorators
class User(Consumer):
    """Class for Users"""
    def __init__(self, session):
        super().__init__(base_url=session._tenant_url, client=session._session)

    @get('/api/web/custom/namespaces/{namespace}/user_roles')
    def list(self, namespace: Path = 'system'):
        """List all Users"""

    @json
    @post('/api/web/custom/namespaces/{namespace}/user_roles')
    def create(self, payload: Body, namespace: Path = 'system'):
        """
        Create a User
        Use create_payload() to build Body
        """

    @json
    @post('/api/web/custom/namespaces/{namespace}/users/cascade_delete')
    def delete(self, payload: Body, namespace: Path = 'system'):
        """
        Delete a User
        use delete_payload() to build Body
        """

    @json
    @put('/api/web/custom/namespaces/{namespace}/users/group_add')
    def group(self, payload: Body, namespace: Path = 'system'):
        """
        Configure Group(s) for a user
        Use group_payload() to build Body
        """

    @staticmethod
    def create_payload(
            email: str,
            first_name: str,
            last_name: str,
            group_names: list=None,
            namespace_roles: list=None,
            idm_type: str = 'SSO',
            namespace: str = 'system'
        ):
        """Payload for Create"""
        if group_names is None:
            group_names = []
        if namespace_roles is None:
            namespace_roles = []
        return {
            "email": email,
            "first_name": first_name,
            "group_names": group_names,
            "idm_type": idm_type,
            "last_name": last_name,
            "name": email,
            "namespace": namespace,
            "namespace_roles": namespace_roles,
            "type": "USER"
        }

    @staticmethod
    def delete_payload(email: str, namespace: str = 'system'):
        """Payload for delete"""
        return {
            "email": email,
            "namespace": namespace
        }

    @staticmethod
    def group_payload(user: str, group_names: list):
        """
        Payload for group modifications
        *group_names is being replaced
        """
        return {
            "group_names": group_names,
            "username": user
        }



"""Module for Groups"""
from uplink import Consumer, Path, Body, json, get, post, put # pylint: disable=unused-import
from . import helper


@helper.common_decorators
class Group(Consumer):
    """Class for Groups"""
    def __init__(self, session):
        super().__init__(base_url=session._tenant_url, client=session._session)

    @get('/api/web/custom/namespaces/{namespace}/user_groups')
    def list(self, namespace: Path = 'system'):
        """List all Groups"""

    @get('/api/web/namespaces/{namespace}/user_groups/{name}')
    def get(self, name: Path, namespace: Path = 'system'):
        """Get a single Group"""

    @json
    @post('/api/web/custom/namespaces/{namespace}/user_groups')
    def create(self, payload: Body, namespace: Path = 'system'):
        """
        Create a Group
        Use create_payload() to build Body
        """

    @json
    @post('/api/web/custom/namespaces/{namespace}/user_groups/{name}')
    def delete(self, name: Path, namespace: Path = 'system'):
        """Delete a group"""

    @json
    @put('/api/web/custom/namespaces/{namespace}/user_groups/{name}/assign_namespace_roles')
    def role_assign(self, payload: Body, name: Path, namespace: Path = 'system'):
        """
        Assign roles for a group
        Use role_payload() to build Body
        """

    @json
    @put('/api/web/custom/namespaces/{namespace}/user_groups/{name}/remove_namespace_roles')
    def role_remove(self, payload: Body, name: Path, namespace: Path = 'system'):
        """
        Remove roles for a group
        Use role_payload() to build Body
        """

    @staticmethod
    def create_payload(
            name: str,
            description: str,
            display_name: str,
            namespace_roles: list,
            usernames: list,
        ):
        """Payload for Create"""
        return {
            "description": description,
            "display_name": display_name,
            "name": name,
            "namespace_roles": namespace_roles,
            "sync_id": "string",
            "usernames": usernames
        }

    @staticmethod
    def role_payload(
            name: str,
            namespace_roles: list
        ):
        """
        Payload for role modifications
        *namespace_roles are added/deleted based on method
        """
        return {
            "name": name,
            "namespace_roles": namespace_roles
        }



"""Package helpers"""
import sys
from datetime import datetime
from uplink import retry, ratelimit, response_handler, error_handler

possible_keys = ['items', 'last_login_map', 'login_events_map', 'users']

def common_decorators(cls):
    """Function to package all uplink decorators for reuse"""
    decorators = [
        #returns.json,
        ratelimit(calls=50, period=50),
        xc_error_handler,
        xc_response_handler,
        xc_extract_items,
        retry(
            when=retry.when.status(503) | retry.when.status(429),
            stop=retry.stop.after_attempt(5) | retry.stop.after_delay(20),
            backoff=retry.backoff.jittered(multiplier=2)
        )
    ]
    for decorator in decorators:
        cls = decorator(cls)
    return cls

def tops_handler(exc_type, exc_value, tb, debug: bool = False):
    """Custom exception handler to give concise errors"""
    if debug:
        sys.__excepthook__(exc_type, exc_value, tb)
    else:
        print(f"{exc_type.__name__}: {exc_value}")

sys.excepthook = tops_handler

class TopsXCException(Exception):
    """Class to where all exceptions should rise"""

@response_handler
def xc_extract_items(json_data: dict):
    """Function to extract possible keys from response"""
    for key in possible_keys:
        if key in json_data:
            return json_data[key]
    return json_data

@response_handler
def xc_response_handler(response):
    """Function to handle HTTP responses"""
    if 200 <= response.status_code < 300:
        try:
            return response.json()
        except Exception as e:
            raise TopsXCException(f"Response not JSON: {str(e)}") from e
    try:
        error_data = response.json()
        error_message = error_data.get("message", "Unknown error occurred")
    except Exception:
        error_message = response.text
    raise TopsXCException(f"API ResponseCode {response.status_code}: {error_message}")

@error_handler(requires_consumer=True)
def xc_error_handler(consumer):
    """Function to handle HTTP client errors"""
    if isinstance(consumer.exceptions.BaseClientException):
        raise TopsXCException("BaseClientException")
    if isinstance(consumer.exceptions.ConnectionError):
        raise TopsXCException("ConnectionError")
    if isinstance(consumer.exceptions.ConnectionTimeout):
        raise TopsXCException("ConnectionTimeout")
    if isinstance(consumer.exceptions.ServerTimeout):
        raise TopsXCException("ServerTimeout")
    if isinstance(consumer.exceptions.SSLError):
        raise TopsXCException("SSLError")
    if isinstance(consumer.exceptions.InvalidURL):
        raise TopsXCException("InvalidURL")

def xc_filter_items(d: dict, keys: list) -> dict:
    """Fuction to filter XC reponse with 'items' dict"""
    items = d.get('items', [])
    filtered_items = [{key: d[key] for key in keys if key in d} for d in items]
    return {'items': filtered_items}

def xc_format_date(date_obj: datetime):
    """
    Function to format dates to what the console expects
    """
    return date_obj.strftime("%Y-%m-%dT%H:%M:%SZ")
